#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Var', 'VarStatus', '_VarBase']

from numvalue import *
from plugin import ComponentRegistration
import types
from set_types import *
from sets import _BaseSet
import pyutilib.math
import pyutilib.misc
from pyutilib.enum import Enum
import sys

VarStatus = Enum('undefined', 'fixed_by_presolve', 'fixed_by_optimizer', 'unused', 'used')

class _VarValue(NumericValue):
    """Holds the numeric value of a variable, along with suffix info"""

    """Constructor"""
    def __init__(self,**kwds):
        NumericValue.__init__(self,**kwds)
        self.var = None # the "parent" variable.
        self.id = None
        self.label = None
        self.initial = None 
        self.lb = None 
        self.ub = None 
        self.fixed = False 
        self.status = VarStatus.undefined 

    def __str__(self):
        return self.name

    def fixed_value(self):
        if self.fixed:
            return True
        return False

    def simplify(self, model):
        return self                 #pragma:nocover

    def pprint(self, ostream=None):
        if ostream is None:
           ostream = sys.stdout
        print >>ostream, str(self),

#
# Variable attributes:
#
# Single variable:
#
# x = Var()
# x.fixed = True
# x.domain = Reals
# x.name = "x"
# x.lb = -1.0
# x.ub = 1.0
# x.initial = 0.0
# x = -0.4
#
# Array of variables:
#
# y = Var(set1,set2,...,setn)
# y.fixed = True (fixes all variables)
# y.domain = Reals
# y.name = "y"
# y[i,j,...,k] (returns value of a given index)
#
# y[i,j,...,k].fixed
# y[i,j,...,k].lb
# y[i,j,...,k].ub
# y[i,j,...,k].initial
#

class _VarBase(object):
    """A numeric variable, which may be defined over a index"""

    """ Constructor
        Arguments:
           name                The name of this variable
           index        The index set that defines the distinct variables.
                          By default, this is None, indicating that there
                          is a single variable.
           domain       A set that defines the type of values that
                          each parameter must be.
           default      A set that defines default values for this
                          variable.
           bounds       A rule for defining bounds values for this
                          variable.
           rule         A rule for setting up this parameter with
                          existing model data
    """
    def __init__(self, **kwd):
        #
        # Default keyword values
        #
        self._attr_declarations = {}
        tmpname="unknown"
        tmpdomain=Reals
        self._ndim=0
        tmpbounds=None
        self._varval={}
        self._initialize=None
        for key in kwd.keys():
          if key == "name":
             tmpname=kwd[key]
          elif key == "within" or key=="domain":
             tmpdomain=kwd[key]
          elif key == "initialize":
             self.__dict__["_"+key] = kwd[key]
          elif key == "bounds":
             tmpbounds=kwd[key]
          else:
             raise ValueError, "Var constructor: unknown keyword - "+key
        self.name = tmpname
        self.domain = tmpdomain
        self.bounds=tmpbounds
        self._index_set=None
        self._model=None

    def as_numeric(self):
        if None in self._varval:
            return self._varval[None]
        return self

    def keys(self):
        return self._varval.keys()

    def __iter__(self):
        return self._varval.keys().__iter__()

    def __contains__(self,ndx):
        return ndx in self._varval

    def dim(self):
        return self._ndim

    def reset(self):
        for key in self._varval:
            if not self._varval[key].initial is None:
                self._varval[key].set_value( self._varval[key].initial )

    def __len__(self):
        return len(self._varval)

    def __setitem__(self,ndx,val):
        #print "HERE",ndx,val, self._valid_value(val,False), self.domain
        if None in self._index:
            raise KeyError, "Cannot set an array value in singleton variable "+self.name
        if ndx not in self._index:
            raise KeyError, "Cannot set the value of array variable "+self.name+" with invalid index "+str(ndx)
        if not self._valid_value(val,False):
            raise ValueError, "Cannot set variable "+self.name+" with invalid value: index=" + str(ndx) + " value=" + str(val)
        self._varval[ndx].value = val

    def __getitem__(self,ndx):
        """This method returns a _VarValue object.  This object can be
           coerced to a numeric value using the value() function, or using
           explicity coercion with float().
        """
        try:
            return self._varval[ndx]
        except KeyError:
            raise KeyError, "Unknown index " + str(ndx) + " in variable " + self.name

    def X__setattr__(self,name,val):
        if name == "_attr_declarations":
            NumericValue.__setattr__(self,name,val)
        if (name in ("value","initial","ub","lb","fixed","status") or name in self._attr_declarations) and len(self._varval)==0:
           raise TypeError, "Attempting to set option '"+name+"' with uninitialized variable "+self.name
        if name == "value":
           tmpval = value(val)
           if tmpval is None:
              raise ValueError, "Value "+str(val)+" is not a valid value for variable "+self.name
           for key in self._varval:
             self._varval[key].value = tmpval
           try:
              self._valid_value(tmpval)
           except ValueError, err:
              raise ValueError, "Bad value for variable "+self.name+" : "+str(err)
        elif name == "initial":
           tmpval = value(val)
           if tmpval is None:
              raise ValueError, "Value "+str(val)+" is not a valid value for variable "+self.name
           for key in self._varval:
             self._varval[key].value = tmpval
             self._varval[key].initial = tmpval
           self._valid_value(tmpval)
        elif name == "ub" or name == "lb":
           tmpval = value(val)
           if tmpval is None:
              raise ValueError, "Value "+str(val)+" is not a valid bound value for variable "+self.name
           for key in self._varval:
             if name == "ub":
                self._varval[key].ub = tmpval
             else:
                self._varval[key].lb = tmpval
           self._valid_value(tmpval)
        elif name == "fixed":
           if type(val) is not bool:
              raise ValueError, "Must use a boolean to specify the 'fixed' attribute for variable "+self.name
           for key in self._varval:
             self._varval[key].fixed = val
        elif name == "status":
           if not val in VarStatus:
              raise ValueError, "Must use a VarStatus value to specify the 'status' attribute for variable "+self.name
           for key in self._varval:
             self._varval[key].status = val
        elif name in self._attr_declarations:
            for key in self._varval:
                setattr(self._varval[key],name,val)
        else:
            NumericValue.__setattr__(self,name,val)

    def X__getattr__(self,name):
        if name == "_attr_declarations":
            return NumericValue.__getattr__(self,name)
        if name in ("value","ub","lb","fixed","initial","status") or name in self._attr_declarations:
           if None not in self._varval:
              raise AttributeError, "Cannot retrieve attribute '"+name+"' for array variable "+self.name
           return getattr(self._varval[None], name)
        return NumericValue.__getattr__(self,name)

    def simplify(self, model):
        return self

    def _construct(self, model, data):
        self._model = model
        #
        # Construct _VarValue() objects for all index values
        #
        if self._ndim > 0:
            if type(self._initialize) is dict:
                self._index = self._initialize.keys()
            for ndx in self._index:
                if ndx is not None:
                    self._varval[ndx] = _VarValue(name=self.name+"["+str(ndx)+"]",domain=self.domain)
                    self._varval[ndx].var = self
                    
        #
        # Initialize values with a dictionary if provided
        #
        if self._initialize is not None and type(self._initialize) is not types.FunctionType:
           if type(self._initialize) is dict:
              for key in self._initialize:
                self._varval[key].value = None
                self._varval[key].initial = self._initialize[key]
                self._valid_value(self._initialize[key],True)
           else:
              for key in self._index:
                self._varval[key].value = None
                self._varval[key].initial = self._initialize
              self._valid_value(self._initialize,True)
        #
        # Initialize values with the _rule function if provided
        #
        elif type(self._initialize) is types.FunctionType:
           for key in self._varval:
             if isinstance(key,tuple):
                tmp = list(key)
             elif key is None:
                tmp = []
             else:
                tmp = [key]
             tmp.append(model)
             tmp = tuple(tmp)
             self._varval[key].value = None
             self._varval[key].initial = self._initialize(*tmp)
             self._valid_value(self._varval[key].initial,True)
        #
        # Initialize bounds with the bounds function if provided
        #
        if self.bounds is not None:
           if type(self.bounds) is tuple:
             for key in self._varval:
               (self._varval[key].lb, self._varval[key].ub) = self.bounds
           else:
             for key in self._varval:
               if isinstance(key,tuple):
                  tmp = list(key)
               elif key is None:
                  tmp = []
               else:
                  tmp = [key]
               tmp.append(model)
               tmp = tuple(tmp)
               (self._varval[key].lb, self._varval[key].ub) = self.bounds(*tmp)
           for key in self._varval:
             if self._varval[key].lb is not None and \
                not pyutilib.math.is_finite(self._varval[key].lb):
                self._varval[key].lb = None
             if self._varval[key].ub is not None and \
                not pyutilib.math.is_finite(self._varval[key].ub):
                self._varval[key].ub = None
        #
        # Iterate through all variables, and tighten the bounds based on
        # the domain bounds information.
        #
        dbounds = self.domain.bounds()
        if not dbounds is None and dbounds != (None,None):
            for key in self._varval:
                if not dbounds[0] is None:
                    if self._varval[key].lb is None or dbounds[0] > self._varval[key].lb:
                        self._varval[key].lb = dbounds[0]
                if not dbounds[1] is None:
                    if self._varval[key].ub is None or dbounds[1] < self._varval[key].ub:
                        self._varval[key].ub = dbounds[1]
        #
        # Setup declared attributes for all variables
        #
        for attr in self._attr_declarations:
            for key in self._varval:
                setattr(self._varval[key],attr,self._attr_declarations[attr][0])

    def pprint(self, ostream=None):
        if ostream is None:
           ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        print >>ostream, "\tSize="+str(len(self)),
        print >>ostream, "\tDomain="+self.domain.name
        if self._index_set is not None:
            print >>ostream, "\tIndicies: ",
            for idx in self._index_set:
               print >>ostream, str(idx.name)+", ",
            print ""
        if None in self._varval:
           print >>ostream, "\tInitial Value : Lower Bound : Upper Bound : Current Value: Fixed"
           print >>ostream, "\t", str(self._varval[None].initial)+" : "+str(self._varval[None].lb)+" : "+str(self._varval[None].ub)+" : "+str(self._varval[None].value)+" : "+str(self._varval[None].fixed)
        else:
           print >>ostream, "\tKey : Initial Value : Lower Bound : Upper Bound : Current Value: Fixed"
           tmp=self._varval.keys()
           tmp.sort()
           for key in tmp:
             val = self._varval[key].initial
             print >>ostream, "\t"+str(key)+" : "+str(val)+" : "+str(value(self._varval[key].lb))+" : "+str(value(self._varval[key].ub))+" : "+str(value(self._varval[key].value))+" : "+str(value(self._varval[key].fixed))

    def display(self, prefix="", ostream=None):
        if ostream is None:
           ostream = sys.stdout
        print >>ostream, prefix+"Variable "+self.name,":",
        print >>ostream, "  Size="+str(len(self)),
        print >>ostream, "Domain="+self.domain.name
        if None in self._varval:
           print >>ostream, prefix+"  Value="+pyutilib.misc.format_io(self._varval[None].value)
        else:
           for key in self._varval:
             val = self._varval[key].value
             print >>ostream, prefix+"  "+str(key)+" : "+str(val)

    def declare_attribute(self, name, default=None):
        """
        Declare a user-defined attribute.
        """
        if name[0] == "_":
            raise AttributeError, "Cannot define an attribute that begins with  '_'"
        if name in self._attr_declarations:
            raise AttributeError, "Attribute %s is already defined" % name
        self._attr_declarations[name] = (default,)
        #if not default is None:
            #self._valid_value(default)
        #
        # If this variable has been constructed, then
        # generate this attribute for all variables
        #
        if len(self._varval) > 0:
            for key in self._varval:
                setattr(self._varval[key],name,default)


class _VarElement(_VarBase,_VarValue):

    def __init__(self, *args, **kwd):
        _VarValue.__init__(self, **kwd)
        _VarBase.__init__(self, **kwd)
        self._index=[None]
        self._varval[None] = self
        self._varval[None].var = self


class _VarArray(_VarBase):

    def __init__(self, *args, **kwd):
        _VarBase.__init__(self, **kwd)
        for arg in args:
            if not isinstance(arg,_BaseSet):
                raise ValueError, "Cannot index variable `"+self.name+"' with object of type "+str(type(arg))
            self._ndim += arg.dimen
        if len(args) == 1:
            self._index=args[0]
        else:
            self._index=None
            self._index_set=args
        #print "X",kwd
        self._dummy_val = _VarValue(**kwd)

    def __float__(self):
        raise TypeError, "Cannot access the value of array variable "+self.name

    def __int__(self):
        raise TypeError, "Cannot access the value of array variable "+self.name

    def _valid_value(self,value,use_exception=True):
        #print "VALID",self._dummy_val.domain
        return self._dummy_val._valid_value(value,use_exception)

    def set_value(self, value):
        raise ValueError, "Cannot specify the value of array variable "+self.name

    def __str__(self):
        return self.name

class Var(object):
    """
    Variable objects that are used to construct Pyomo models.
    """
    def __new__(cls, *args, **kwds):
        if args == ():
            self = _VarElement(*args, **kwds)
        else:
            self = _VarArray(*args, **kwds)
        return self


ComponentRegistration("Var", _VarBase, "Decision variables in a model.")
